import const
import event
import export


class Fragment(object):

    def __init__(self, event_list, onset=0.0):

        #: List. A list of Note and/or Chord objects.
        self.event_list = event_list

        self._onset = None
        #: Float. The time of onset. The unit of measurement is abstract, so
        #: the user may interpret *onset* however they wish, e.g. absolute
        #: time, musical time, etc.
        self.onset = onset

    @property
    def onset(self):
        return self._onset

    @onset.setter
    def onset(self, value):
        self._onset = value
        self.queue()

    # TODO: magic methods

    def __repr__(self):
        s = ""
        for e in self.event_list:
            s += e.__repr__()
        return s

    # data
    # ------------------------------------------------------------------- #

    def flat(self):
        """
        Return flattened *event_list*. WARNING -- operating on this list may
        result in unexpected behaviour.

        :returns: list -- list of Note objects

        """
        chord_list = [x for x in self.event_list if isinstance(x, event.Chord)]
        note_list = [x for x in self.event_list if isinstance(x, event.Note)]
        for c in chord_list:
            note_list += c.note_list
        return note_list

    def onset_list(self):  # TODO: generalize (see pitch_list)
        result = []
        for e in self.event_list:
            if isinstance(e, event.Chord):
                for n in e:
                    result.append(n.onset)
            if isinstance(e, event.Note):
                result.append(e.onset)
        return result

    def pitch_list(self):  # TODO: generalize (see onset_list)
        result = []
        for e in self.event_list:
            if isinstance(e, event.Chord):
                for n in e:
                    result.append(n.pitch)
            if isinstance(e, event.Note):
                result.append(e.pitch)
        return result

    # info
    # ------------------------------------------------------------------- #

    def ambitus(self):
        """
        Return interval between highest and lowest events in *event_list*.

        :returns: int -- interval

        """
        return self.limit(const.UPPER).pitch - self.limit(const.LOWER).pitch

    def duration(self):
        """
        Return duration of Fragment object.

        :returns: float -- duration

        """
        duration = 0.0
        for e in self.event_list:
            duration += e.duration
        return duration

    def limit(self, limit):
        """
        Return highest or lowest Note object as specified by *limit*, a
        predefined constant.

        :param limit: the limit to search (lower or upper)
        :type limit: int
        :returns: Note

        """
        result = None
        # separate chord and note
        chord_list = [x for x in self.event_list if isinstance(x, event.Chord)]
        note_list = [x for x in self.event_list if isinstance(x, event.Note)]
        # extract limit from chord
        if chord_list:
            for c in chord_list:
                if limit == const.LOWER:
                    note_list.append(c.limit(const.LOWER))
                if limit == const.UPPER:
                    note_list.append(c.limit(const.UPPER))
        # find limit
        if note_list:
            result = note_list[0]
            for n in note_list:
                if limit == const.LOWER:
                    if n.pitch < result.pitch:
                        result = n
                if limit == const.UPPER:
                    if n.pitch > result.pitch:
                        result = n
        return result

    # round
    # ------------------------------------------------------------------- #

    def quantize(self):
        """
        Quantize event objects in *event_list*.

        """
        for e in self.event_list:
            e.quantize()

    def tune(self):
        """
        Tune event objects in *event_list*.

        """
        for e in self.event_list:
            e.tune()

    # sort
    # ------------------------------------------------------------------- #

    def queue(self):
        """
        Queue event objects in *event_list* relative to *onset*.

        """
        o = self.onset
        for e in self.event_list:
            e.onset = o
            o += e.duration

    def sort(self):
        """
        Sort *event_list* by *onset*.

        """
        self.event_list.sort(key=lambda x: x.onset)

    # transform
    # ------------------------------------------------------------------- #

    def augment(self, factor):
        """
        Multiply *duration* by *factor* (i.e. scale on x axis). If *factor* is
        less than 1.0 but greater than 0.0, the effect will be diminution.

        :param factor: the factor of augmentation
        :type factor: float

        """
        for e in self.event_list:
            e.augment(factor)
        self.queue()

    def dilate(self, factor):  # TODO: direction of dilation
        """
        Dilate a Fragment object by *factor* (i.e. scale on y axis). The lowest
        Note object will remain constant. If *factor* is less than 1.0 but
        greater than 0.0, the effect will be contraction. If *factor* is
        negative, the dilation will be reflected.

        :param factor: the factor of dilation
        :type factor: float

        """
        lowest = self.limit(const.LOWER)
        delta = (lowest.pitch * factor) - lowest.pitch
        for e in self.event_list:
            if isinstance(e, event.Chord):
                e.dilate(factor, lowest)
            if isinstance(e, event.Note):
                e.pitch *= factor
                e.pitch -= delta
        self.tune()

    def invert(self, note):
        """
        Invert a Fragment object about *note* (i.e. reflect on x axis).

        :param note: the axis of inversion
        :type note: Note

        """
        for e in self.event_list:
            e.invert(note)
        self.tune()

    def retrograde(self):
        """
        Reverse order of event objects in *event_list* (i.e. reflect on y axis).

        """
        for e in self.event_list:
            e.onset += e.duration
            e.onset *= -1
        self.sort()
        self.queue()

    def rotate(self, n):
        """
        Rotate *event_list* by *n* places. Positive values for *n* rotate
        right, whereas negative values for *n* rotate left.

        :param n: the number of places to rotate
        :type n: int

        """
        length = len(self.event_list)
        self.event_list = self.event_list[-n % length:] + \
            self.event_list[:-n % length]
        self.queue()

    def shift(self, duration):
        """
        Shift the time of onset by *duration* (i.e. translate on x axis).

        :param duration: a duration (positive or negative)
        :type duration: float

        """
        self.onset += duration

    def transpose(self, interval):
        """
        Transpose a Fragment object by *interval* (i.e translate on y axis).

        :param interval: an interval in semitones (positive or negative)
        :type interval: int

        """
        for e in self.event_list:
            e.transpose(interval)


class Part(object):

    def __init__(self, fragment_list, instrument=0):

        self._fragment_list = None
        #: List. A list of Fragment objects.
        self.fragment_list = fragment_list

        #: Integer. General MIDI instrument number, aka program change number.
        #: Default is 0 (Acoustic Grand Piano). See const.py for the General
        #: MIDI Level 1 Instrument Patch Map.
        self.instrument = instrument

        #: Integer (range 0 - 127). The MIDI track. Set by Score object.
        self.track = None

        #: Integer (range 0 - 15). The MIDI channel. The default is 0.
        self.channel = 0

    @property
    def fragment_list(self):
        return self._fragment_list

    @fragment_list.setter
    def fragment_list(self, value):
        self._fragment_list = value
        self.queue()

    # TODO: magic methods

    def __repr__(self):
        s = ""
        for f in self.fragment_list:
            s += f.__repr__()
            if f is not self.fragment_list[-1]:
                s += "\n"
        return s

    # convert
    # ------------------------------------------------------------------- #

    def midi_to_instrument(self):
        instrument_name = None
        for k, v in const.GENERAL_MIDI.iteritems():
            if v == self.instrument:
                instrument_name = k
                break
        return instrument_name

    # sort
    # ------------------------------------------------------------------- #

    def queue(self):
        """
        Queue Fragment objects in *fragment_list*.

        """
        o = self.fragment_list[0].onset
        for f in self.fragment_list:
            f.onset = o
            o += f.duration()


class Score(object):

    def __init__(self, part_list, tempo=60):

        self._part_list = None
        #: List. A list of Part objects.
        self.part_list = part_list

        #: Integer. The tempo in beats per minute (BPM).
        self.tempo = tempo

    @property
    def part_list(self):
        return self._part_list

    @part_list.setter
    def part_list(self, value):
        self._part_list = value
        # set track number for part
        for i in range(len(self.part_list)):
            part = self.part_list[i]
            part.track = i

    # TODO: magic methods

    def __repr__(self):
        s = ""
        for p in self.part_list:
            s += p.midi_to_instrument()
            s += "\n"
            s += p.__repr__()
        return s

    # export
    # ------------------------------------------------------------------- #

    def export(self, path=None):
        """
        Export to a MIDI file. For ease of path construction, use Python's
        os.path module. If *path* is None, the file will be written to the
        user's home directory.

        :param path: the export path
        :type path: string
        """
        export.midi(self.part_list, self.tempo, path)
